@@ -11,7 +11,7 @@ import sys
 import uuid
 from logging import getLogger
 from threading import Lock, Timer
-from typing import Optional
+from typing import IO, Dict, List, Optional, Tuple, Union
 
 from .compat import BASE_EXCEPTION_CLASS
 from .constants import FIELD_NAME_TO_ID, PARAMETER_PYTHON_CONNECTOR_QUERY_RESULT_FORMAT
@@ -33,8 +33,9 @@ from .telemetry import TelemetryData, TelemetryField
 from .time_util import get_time_millis
 
 MYPY = False
-if MYPY:  # from typing import TYPE_CHECKING once 3.5 is deprecated
+if MYPY:  # pragma: no cover
     from .connection import SnowflakeConnection
+    from .file_transfer_agent import SnowflakeProgressPercentage
 
 logger = getLogger(__name__)
 
@@ -47,7 +48,7 @@ except ImportError:
 try:
     from .arrow_result import ArrowResult
     CAN_USE_ARROW_RESULT = True
-except ImportError as e:
+except ImportError as e:  # pragma: no cover
     logger.debug("Failed to import ArrowResult. No Apache Arrow result set format can be used. ImportError: %s", e)
     CAN_USE_ARROW_RESULT = False
 
@@ -70,7 +71,7 @@ DESC_TABLE_RE = re.compile(r'desc(?:ribe)?\s+([\w_]+)\s*;?\s*$',
 LOG_MAX_QUERY_LENGTH = 80
 
 
-def exit_handler(*_):
+def exit_handler(*_):  # pragma: no cover
     """Handler for signal. When called, it will raise SystemExit with exit code FORCE_EXIT."""
     print("\nForce exit")
     logger.info("Force exit")
@@ -162,7 +163,7 @@ class SnowflakeCursor(object):
 
         self.reset()
 
-    def __del__(self):
+    def __del__(self):  # pragma: no cover
         try:
             self.close()
         except BASE_EXCEPTION_CLASS as e:
@@ -265,8 +266,11 @@ class SnowflakeCursor(object):
                 'errno': ER_UNSUPPORTED_METHOD,
                 'sqlstate': SQLSTATE_FEATURE_NOT_SUPPORTED})
 
-    def close(self):
-        """Closes the cursor object."""
+    def close(self) -> Optional[bool]:
+        """Closes the cursor object.
+
+        Returns whether the cursor was closed during this call.
+        """
         try:
             if self.is_closed():
                 return False
@@ -282,10 +286,14 @@ class SnowflakeCursor(object):
     def is_closed(self):
         return self._connection is None or self._connection.is_closed()
 
-    def _execute_helper(
-            self, query, timeout=0, statement_params=None,
-            binding_params=None,
-            is_internal=False, _no_results=False, _is_put_get=None):
+    def _execute_helper(self,
+                        query: str,
+                        timeout: int = 0,
+                        statement_params: Optional[Dict[str, str]] = None,
+                        binding_params: Union[Tuple, Dict[str, Dict[str, str]]] = None,
+                        is_internal: bool = False,
+                        _no_results: bool = False,
+                        _is_put_get=None):
         del self.messages[:]
 
         if statement_params is not None and not isinstance(
@@ -294,8 +302,7 @@ class SnowflakeCursor(object):
                 self.connection, self,
                 ProgrammingError,
                 {
-                    'msg': "The data type of statement params is invalid. "
-                            "It must be dict.",
+                    'msg': "The data type of statement params is invalid. It must be dict.",
                     'errno': ER_INVALID_VALUE,
                 })
 
@@ -341,7 +348,7 @@ class SnowflakeCursor(object):
 
         original_sigint = signal.getsignal(signal.SIGINT)
 
-        def interrupt_handler(*_):
+        def interrupt_handler(*_):  # pragma: no cover
             try:
                 signal.signal(signal.SIGINT, exit_handler)
             except (ValueError, TypeError):
@@ -365,7 +372,7 @@ class SnowflakeCursor(object):
         try:
             if not original_sigint == exit_handler:
                 signal.signal(signal.SIGINT, interrupt_handler)
-        except ValueError:
+        except ValueError:  # pragma: no cover
             logger.debug(
                 'Failed to set SIGINT handler. '
                 'Not in main thread. Ignored...')
@@ -376,7 +383,7 @@ class SnowflakeCursor(object):
                 self._sequence_counter,
                 self._request_id,
                 binding_params=binding_params,
-                is_file_transfer=self._is_file_transfer,
+                is_file_transfer=bool(self._is_file_transfer),
                 statement_params=statement_params,
                 is_internal=is_internal,
                 _no_results=_no_results)
@@ -384,7 +391,7 @@ class SnowflakeCursor(object):
             try:
                 if original_sigint:
                     signal.signal(signal.SIGINT, original_sigint)
-            except (ValueError, TypeError):
+            except (ValueError, TypeError):  # pragma: no cover
                 logger.debug(
                     'Failed to reset SIGINT handler. Not in main '
                     'thread. Ignored...')
@@ -420,19 +427,19 @@ class SnowflakeCursor(object):
         return ret
 
     def execute(self,
-                command,
-                params=None,
+                command: str,
+                params: Union[List, Tuple, None] = None,
                 timeout: Optional[int] = None,
                 exec_async: bool = False,
                 _do_reset: bool = True,
-                _put_callback=None,
-                _put_azure_callback=None,
-                _put_callback_output_stream=sys.stdout,
-                _get_callback=None,
-                _get_azure_callback=None,
-                _get_callback_output_stream=sys.stdout,
+                _put_callback: 'SnowflakeProgressPercentage' = None,
+                _put_azure_callback: 'SnowflakeProgressPercentage' = None,
+                _put_callback_output_stream: IO[str] = sys.stdout,
+                _get_callback: 'SnowflakeProgressPercentage' = None,
+                _get_azure_callback: 'SnowflakeProgressPercentage' = None,
+                _get_callback_output_stream: IO[str] = sys.stdout,
                 _show_progress_bar: bool = True,
-                _statement_params=None,
+                _statement_params: Optional[Dict[str, str]] = None,
                 _is_internal: bool = False,
                 _no_results: bool = False,
                 _use_ijson: bool = False,
@@ -499,8 +506,8 @@ class SnowflakeCursor(object):
                 # qmark and numeric paramstyle
                 # server side binding
                 query = command
-                processed_params = self._connection._process_params_qmarks(
-                    params, self)
+                # TODO we could probably rework this to not make dicts like this: {'1': 'value', '2': '13'}
+                processed_params = self._connection._process_params_qmarks(params, self)
         # Skip reporting Key, Value and Type errors
         except (KeyError, ValueError, TypeError):
             raise
@@ -823,8 +830,7 @@ class SnowflakeCursor(object):
         if size < 0:
             errorvalue = {
                 'msg': ("The number of rows is not zero or "
-                         "positive number: {0}").format(
-                    size),
+                         "positive number: {}").format(size),
                 'errno': ER_NOT_POSITIVE_SIZE}
             Error.errorhandler_wrapper(
                 self.connection, self, ProgrammingError, errorvalue)
