@@ -19,7 +19,7 @@ from datetime import datetime, timedelta
 from logging import getLogger
 from os import environ, path
 from os.path import expanduser
-from threading import Lock
+from threading import Lock, RLock
 from time import gmtime, strftime
 
 import jwt
@@ -28,12 +28,25 @@ import requests as generic_requests
 from snowflake.connector.compat import OK, urlsplit
 from snowflake.connector.constants import HTTP_HEADER_USER_AGENT
 from snowflake.connector.errorcode import (
-    ER_INVALID_OCSP_RESPONSE,
-    ER_INVALID_OCSP_RESPONSE_CODE,
+    ER_INVALID_OCSP_RESPONSE_SSD,
     ER_INVALID_SSD,
-    ER_OCSP_FAILED_TO_CONNECT_HOST,
-    ER_SERVER_CERTIFICATE_REVOKED,
-    ER_SERVER_CERTIFICATE_UNKNOWN,
+    ER_OCSP_FAILED_TO_CONNECT_CACHE_SERVER,
+    ER_OCSP_RESPONSE_ATTACHED_CERT_EXPIRED,
+    ER_OCSP_RESPONSE_ATTACHED_CERT_INVALID,
+    ER_OCSP_RESPONSE_CACHE_DECODE_FAILED,
+    ER_OCSP_RESPONSE_CACHE_DOWNLOAD_FAILED,
+    ER_OCSP_RESPONSE_CERT_STATUS_INVALID,
+    ER_OCSP_RESPONSE_CERT_STATUS_REVOKED,
+    ER_OCSP_RESPONSE_CERT_STATUS_UNKNOWN,
+    ER_OCSP_RESPONSE_EXPIRED,
+    ER_OCSP_RESPONSE_FETCH_EXCEPTION,
+    ER_OCSP_RESPONSE_FETCH_FAILURE,
+    ER_OCSP_RESPONSE_INVALID_EXPIRY_INFO_MISSING,
+    ER_OCSP_RESPONSE_INVALID_SIGNATURE,
+    ER_OCSP_RESPONSE_LOAD_FAILURE,
+    ER_OCSP_RESPONSE_STATUS_UNSUCCESSFUL,
+    ER_OCSP_RESPONSE_UNAVAILABLE,
+    ER_OCSP_URL_INFO_MISSING,
 )
 from snowflake.connector.errors import RevocationCheckError
 from snowflake.connector.network import PYTHON_CONNECTOR_USER_AGENT
@@ -51,7 +64,51 @@ logger = getLogger(__name__)
 
 class OCSPTelemetryData(object):
 
+    CERTIFICATE_EXTRACTION_FAILED = "CertificateExtractionFailed"
+    OCSP_URL_MISSING = "OCSPURLMissing"
+    OCSP_RESPONSE_UNAVAILABLE = "OCSPResponseUnavailable"
+    OCSP_RESPONSE_FETCH_EXCEPTION = "OCSPResponseFetchException"
+    OCSP_RESPONSE_FAILED_TO_CONNECT_CACHE_SERVER = "OCSPResponseFailedToConnectCacheServer"
+    OCSP_RESPONSE_CERT_STATUS_INVALID = "OCSPResponseCertStatusInvalid"
+    OCSP_RESPONSE_CERT_STATUS_REVOKED = "OCSPResponseCertStatusRevoked"
+    OCSP_RESPONSE_CERT_STATUS_UNKNOWN = "OCSPResponseCertStatusUnknown"
+    OCSP_RESPONSE_STATUS_UNSUCCESSFUL = "OCSPResponseStatusUnsuccessful"
+    OCSP_RESPONSE_ATTACHED_CERT_INVALID = "OCSPResponseAttachedCertInvalid"
+    OCSP_RESPONSE_ATTACHED_CERT_EXPIRED = "OCSPResponseAttachedCertExpired"
+    OCSP_RESPONSE_INVALID_SIGNATURE = "OCSPResponseSignatureInvalid"
+    OCSP_RESPONSE_EXPIRY_INFO_MISSING = "OCSPResponseExpiryInfoMissing"
+    OCSP_RESPONSE_EXPIRED = "OCSPResponseExpired"
+    OCSP_RESPONSE_FETCH_FAILURE = "OCSPResponseFetchFailure"
+    OCSP_RESPONSE_CACHE_DOWNLOAD_FAILED = "OCSPResponseCacheDownloadFailed"
+    OCSP_RESPONSE_CACHE_DECODE_FAILED = "OCSPResponseCacheDecodeFailed"
+    OCSP_RESPONSE_LOAD_FAILURE = "OCSPResponseLoadFailure"
+    OCSP_RESPONSE_INVALID_SSD = "OCSPResponseInvalidSSD"
+
+    ERROR_CODE_MAP = {
+        ER_OCSP_URL_INFO_MISSING: OCSP_URL_MISSING,
+        ER_OCSP_RESPONSE_UNAVAILABLE: OCSP_RESPONSE_UNAVAILABLE,
+        ER_OCSP_RESPONSE_FETCH_EXCEPTION: OCSP_RESPONSE_FETCH_EXCEPTION,
+        ER_OCSP_FAILED_TO_CONNECT_CACHE_SERVER: OCSP_RESPONSE_FAILED_TO_CONNECT_CACHE_SERVER,
+        ER_OCSP_RESPONSE_CERT_STATUS_INVALID: OCSP_RESPONSE_CERT_STATUS_INVALID,
+        ER_OCSP_RESPONSE_CERT_STATUS_REVOKED: OCSP_RESPONSE_CERT_STATUS_REVOKED,
+        ER_OCSP_RESPONSE_CERT_STATUS_UNKNOWN: OCSP_RESPONSE_CERT_STATUS_UNKNOWN,
+        ER_OCSP_RESPONSE_STATUS_UNSUCCESSFUL: OCSP_RESPONSE_STATUS_UNSUCCESSFUL,
+        ER_OCSP_RESPONSE_ATTACHED_CERT_INVALID: OCSP_RESPONSE_ATTACHED_CERT_INVALID,
+        ER_OCSP_RESPONSE_ATTACHED_CERT_EXPIRED: OCSP_RESPONSE_ATTACHED_CERT_EXPIRED,
+        ER_OCSP_RESPONSE_INVALID_SIGNATURE: OCSP_RESPONSE_INVALID_SIGNATURE,
+        ER_OCSP_RESPONSE_INVALID_EXPIRY_INFO_MISSING: OCSP_RESPONSE_EXPIRY_INFO_MISSING,
+        ER_OCSP_RESPONSE_EXPIRED: OCSP_RESPONSE_EXPIRED,
+        ER_OCSP_RESPONSE_FETCH_FAILURE: OCSP_RESPONSE_FETCH_FAILURE,
+        ER_OCSP_RESPONSE_LOAD_FAILURE: OCSP_RESPONSE_LOAD_FAILURE,
+        ER_OCSP_RESPONSE_CACHE_DOWNLOAD_FAILED: OCSP_RESPONSE_CACHE_DOWNLOAD_FAILED,
+        ER_OCSP_RESPONSE_CACHE_DECODE_FAILED: OCSP_RESPONSE_CACHE_DECODE_FAILED,
+        ER_INVALID_OCSP_RESPONSE_SSD: OCSP_RESPONSE_INVALID_SSD,
+        ER_INVALID_SSD: OCSP_RESPONSE_INVALID_SSD
+    }
+
     def __init__(self):
+        self.event_sub_type = None
+        self.ocsp_connection_method = None
         self.cert_id = None
         self.sfc_peer_host = None
         self.ocsp_url = None
@@ -62,6 +119,26 @@ class OCSPTelemetryData(object):
         self.fail_open = False
         self.insecure_mode = False
 
+    def set_event_sub_type(self, event_sub_type: str) -> None:
+        """
+        Sets sub type for OCSP Telemetry Event.
+
+        There can be multiple event_sub_type that could have happened
+        during a single connection establishment. Ensure that all of them
+        are captured.
+        :param event_sub_type:
+        :return:
+        """
+        if self.event_sub_type is not None:
+            self.event_sub_type = "{}|{}".\
+                format(self.event_sub_type,
+                       event_sub_type)
+        else:
+            self.event_sub_type = event_sub_type
+
+    def set_ocsp_connection_method(self, ocsp_conn_method: str) -> None:
+        self.ocsp_connection_method = ocsp_conn_method
+
     def set_cert_id(self, cert_id):
         self.cert_id = cert_id
 
@@ -95,9 +172,10 @@ class OCSPTelemetryData(object):
         self.insecure_mode = insecure_mode
 
     def generate_telemetry_data(self, event_type, urgent=False):
-        cls, exception, stack_trace = sys.exc_info()
+        _, exception, _ = sys.exc_info()
         telemetry_data = {}
         telemetry_data.update({"eventType": event_type})
+        telemetry_data.update({"eventSubType": self.event_sub_type})
         telemetry_data.update({"sfcPeerHost": self.sfc_peer_host})
         telemetry_data.update({"certId": self.cert_id})
         telemetry_data.update({"ocspRequestBase64": self.ocsp_req})
@@ -244,7 +322,8 @@ class OCSPServer(object):
                 retval = OCSPServer._download_ocsp_response_cache(ocsp,
                                                                   self.CACHE_SERVER_URL)
                 if not retval:
-                    raise RevocationCheckError(msg="OCSP Cache Server Unavailable.")
+                    raise RevocationCheckError(msg="OCSP Cache Server Unavailable.",
+                                               errno=ER_OCSP_RESPONSE_CACHE_DOWNLOAD_FAILED)
                 logger.debug("downloaded OCSP response cache file from %s",
                              self.CACHE_SERVER_URL)
                 logger.debug("# of certificates: %s", len(OCSPCache.CACHE))
@@ -252,6 +331,7 @@ class OCSPServer(object):
                 logger.debug("OCSP Response cache download failed. The client"
                              "will reach out to the OCSP Responder directly for"
                              "any missing OCSP responses %s\n" % rce.msg)
+                raise
 
     @staticmethod
     def _download_ocsp_response_cache(ocsp, url, do_retry=True):
@@ -301,7 +381,6 @@ class OCSPServer(object):
                         max_retry)
                     return False
                 return True
-
         except Exception as e:
             logger.debug("Failed to get OCSP response cache from %s: %s", url,
                          e)
@@ -309,7 +388,7 @@ class OCSPServer(object):
                 msg="Failed to get OCSP Response Cache from {}: {}".format(
                     url,
                     e),
-                errno=ER_OCSP_FAILED_TO_CONNECT_HOST)
+                errno=ER_OCSP_FAILED_TO_CONNECT_CACHE_SERVER)
 
     def generate_get_url(self, ocsp_url, b64data):
         parsed_url = urlsplit(ocsp_url)
@@ -376,6 +455,7 @@ class OCSPCache(object):
         """Deletes the OCSP response cache file if exists."""
         cache_file = path.join(OCSPCache.CACHE_DIR, OCSPCache.OCSP_RESPONSE_CACHE_FILE_NAME)
         if path.exists(cache_file):
+            logger.debug('deleting cache file {}'.format(cache_file))
             os.unlink(cache_file)
 
     @staticmethod
@@ -412,15 +492,14 @@ class OCSPCache(object):
                     ocsp,
                     path.join(parsed_url.netloc, parsed_url.path))
             else:
-                raise Exception(
-                    "Unsupported OCSP URI: %s",
-                    OCSPCache.OCSP_RESPONSE_CACHE_URI)
-        except Exception as e:
+                msg = "Unsupported OCSP URI: {}".format(OCSPCache.OCSP_RESPONSE_CACHE_URI)
+                raise Exception(msg)
+        except (RevocationCheckError, Exception) as rce:
             logger.debug(
                 "Failed to read OCSP response cache file %s: %s, "
                 "No worry. It will validate with OCSP server. "
                 "Ignoring...",
-                OCSPCache.OCSP_RESPONSE_CACHE_URI, e, exc_info=True)
+                OCSPCache.OCSP_RESPONSE_CACHE_URI, rce, exc_info=True)
 
     @staticmethod
     def read_ocsp_response_cache_file(ocsp, filename):
@@ -488,7 +567,7 @@ class OCSPCache(object):
     @staticmethod
     def write_ocsp_response_cache_file(ocsp, filename):
         """Writes OCSP Response Cache."""
-        logger.debug('writing OCSP response cache file')
+        logger.debug('writing OCSP response cache file to {}'.format(filename))
         file_cache_data = {}
         ocsp.encode_ocsp_response_cache(file_cache_data)
         with codecs.open(filename, 'w', encoding='utf-8', errors='ignore') as f:
@@ -631,21 +710,27 @@ class OCSPCache(object):
 
         current_cache takes precedence over previous_cache.
         """
-        OCSPCache.clear_cache()
-        if previous_cache_filename:
-            OCSPCache.read_ocsp_response_cache_file(
-                ocsp, previous_cache_filename)
-        previous_cache = deepcopy(OCSPCache.CACHE)
+        try:
+            OCSPCache.clear_cache()
+            if previous_cache_filename:
+                OCSPCache.read_ocsp_response_cache_file(
+                    ocsp, previous_cache_filename)
+            previous_cache = deepcopy(OCSPCache.CACHE)
 
-        OCSPCache.clear_cache()
-        OCSPCache.read_ocsp_response_cache_file(ocsp, current_cache_filename)
-        current_cache = deepcopy(OCSPCache.CACHE)
+            OCSPCache.clear_cache()
+            OCSPCache.read_ocsp_response_cache_file(ocsp, current_cache_filename)
+            current_cache = deepcopy(OCSPCache.CACHE)
 
-        # overwrite the previous one with the current one
-        previous_cache.update(current_cache)
+            # overwrite the previous one with the current one
+            previous_cache.update(current_cache)
 
-        OCSPCache.CACHE = previous_cache
-        OCSPCache.write_ocsp_response_cache_file(ocsp, output_filename)
+            OCSPCache.CACHE = previous_cache
+            OCSPCache.write_ocsp_response_cache_file(ocsp, output_filename)
+        except RevocationCheckError as rce:
+            # Catch any revocation check error and raise a generic exception
+            # Any other generic exception along with this one should be
+            # handled by the calling function / OCSP Cache Server.
+            raise Exception(rce.msg)
 
     @staticmethod
     def _file_timestamp(filename):
@@ -685,6 +770,7 @@ class OCSPCache(object):
         fname = path.join(parsed_url.netloc, parsed_url.path)
         OCSPCache.lock_cache_file(fname)
         try:
+            logger.debug('deleting cache file, used by tests only {}'.format(fname))
             os.unlink(fname)
         finally:
             OCSPCache.unlock_cache_file(fname)
@@ -810,7 +896,7 @@ class SnowflakeOCSP(object):
     ROOT_CERTIFICATES_DICT = {}  # root certificates
 
     # root certificate cache lock
-    ROOT_CERTIFICATES_DICT_LOCK = Lock()
+    ROOT_CERTIFICATES_DICT_LOCK = RLock()
 
     # ssd cache object
     SSD = SFSsd()
@@ -943,6 +1029,7 @@ class SnowflakeOCSP(object):
         try:
             cert_data = self.extract_certificate_chain(connection)
         except RevocationCheckError:
+            telemetry_data.set_event_sub_type(OCSPTelemetryData.CERTIFICATE_EXTRACTION_FAILED)
             logger.debug(telemetry_data.generate_telemetry_data("RevocationCheckFailure"))
             return None
 
@@ -1081,14 +1168,14 @@ class SnowflakeOCSP(object):
                 logger.debug("using OCSP response cache")
 
             if not ocsp_response:
-                logger.debug('No OCSP URL is found.')
+                telemetry_data.set_event_sub_type(OCSPTelemetryData.OCSP_RESPONSE_UNAVAILABLE)
                 raise RevocationCheckError(msg="Could not retrieve OCSP Response. Cannot perform Revocation Check",
-                                           errno=ER_SERVER_CERTIFICATE_UNKNOWN)
+                                           errno=ER_OCSP_RESPONSE_UNAVAILABLE)
             try:
                 self.process_ocsp_response(issuer, cert_id, ocsp_response)
                 err = None
             except RevocationCheckError as op_er:
-                if SnowflakeOCSP.SSD.ACTIVATE_SSD and op_er.errno == ER_INVALID_OCSP_RESPONSE:
+                if SnowflakeOCSP.SSD.ACTIVATE_SSD and op_er.errno == ER_INVALID_OCSP_RESPONSE_SSD:
                     logger.debug(
                         "Potentially the response is a server side directive")
                     if self.process_ocsp_bypass_directive(ocsp_response,
@@ -1098,6 +1185,7 @@ class SnowflakeOCSP(object):
                         # TODO - Remove this potentially broken OCSP Response / SSD
                         raise op_er
                 else:
+                    telemetry_data.set_event_sub_type(OCSPTelemetryData.ERROR_CODE_MAP[op_er.errno])
                     raise op_er
 
         except RevocationCheckError as rce:
@@ -1114,13 +1202,13 @@ class SnowflakeOCSP(object):
 
     def verify_fail_open(self, ex_obj, telemetry_data):
         if not self.is_enabled_fail_open():
-            if ex_obj.errno is ER_SERVER_CERTIFICATE_REVOKED:
+            if ex_obj.errno is ER_OCSP_RESPONSE_CERT_STATUS_REVOKED:
                 logger.debug(telemetry_data.generate_telemetry_data("RevokedCertificateError", True))
             else:
                 logger.debug(telemetry_data.generate_telemetry_data("RevocationCheckFailure"))
             return ex_obj
         else:
-            if ex_obj.errno is ER_SERVER_CERTIFICATE_REVOKED:
+            if ex_obj.errno is ER_OCSP_RESPONSE_CERT_STATUS_REVOKED:
                 logger.debug(telemetry_data.generate_telemetry_data("RevokedCertificateError", True))
                 return ex_obj
             else:
@@ -1133,6 +1221,8 @@ class SnowflakeOCSP(object):
         results = []
         try:
             self._check_ocsp_response_cache_server(cert_data)
+        except RevocationCheckError as rce:
+            telemetry_data.set_event_sub_type(OCSPTelemetryData.ERROR_CODE_MAP[rce.errno])
         except Exception as ex:
             logger.debug("Caught unknown exception - %s. Continue to validate by direct connection", str(ex))
 
@@ -1152,7 +1242,7 @@ class SnowflakeOCSP(object):
         for issuer, subject in cert_data:
             # check if any OCSP response is NOT in cache
             cert_id, _ = self.create_ocsp_request(issuer, subject)
-            in_cache, cache = SnowflakeOCSP.OCSP_CACHE.find_cache(
+            in_cache, _ = SnowflakeOCSP.OCSP_CACHE.find_cache(
                 self, cert_id, subject)
             if not in_cache:
                 # not found any
@@ -1275,8 +1365,9 @@ class SnowflakeOCSP(object):
         cert_id_enc = self.encode_cert_id_base64(
             self.decode_cert_id_key(cert_id))
         if not ocsp_url:
+            telemetry_data.set_event_sub_type(OCSPTelemetryData.OCSP_URL_MISSING)
             raise RevocationCheckError(msg="No OCSP URL found in cert. Cannot perform Certificate Revocation check",
-                                       errno=ER_SERVER_CERTIFICATE_UNKNOWN)
+                                       errno=ER_OCSP_URL_INFO_MISSING)
         headers = {HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT}
 
         if not SnowflakeOCSP.SSD.ACTIVATE_SSD and \
@@ -1306,6 +1397,7 @@ class SnowflakeOCSP(object):
                                   'ocsp_responder_url': ocsp_url})
             headers['Content-Type'] = 'application/json'
 
+        telemetry_data.set_ocsp_connection_method(actual_method)
         if self.test_mode is not None:
             logger.debug("WARNING - DRIVER IS CONFIGURED IN TESTMODE.")
             test_ocsp_url = os.getenv("SF_TEST_OCSP_URL", None)
@@ -1359,19 +1451,21 @@ class SnowflakeOCSP(object):
                                      "Retrying in %s(s)", sleep_time)
                         time.sleep(sleep_time)
                     else:
+                        telemetry_data.set_event_sub_type(OCSPTelemetryData.OCSP_RESPONSE_FETCH_EXCEPTION)
                         raise RevocationCheckError(
                             msg="Could not fetch OCSP Response from server. Consider"
                                 "checking your whitelists : Exception - {}".format(
                                 str(ex)),
-                            errno=ER_OCSP_FAILED_TO_CONNECT_HOST)
+                            errno=ER_OCSP_RESPONSE_FETCH_EXCEPTION)
             else:
                 logger.error(
                     "Failed to get OCSP response after {} attempt. Consider checking "
                     "for OCSP URLs being blocked".format(max_retry))
+                telemetry_data.set_event_sub_type(OCSPTelemetryData.OCSP_RESPONSE_FETCH_FAILURE)
                 raise RevocationCheckError(
                     msg="Failed to get OCSP response after {} attempt.".format(
                         max_retry),
-                    errno=ER_INVALID_OCSP_RESPONSE_CODE)
+                    errno=ER_OCSP_RESPONSE_FETCH_FAILURE)
 
         return ret
 
@@ -1386,7 +1480,7 @@ class SnowflakeOCSP(object):
                 msg="Either this update or next "
                     "update is None. this_update: {}, next_update: {}".format(
                     this_update_native, next_update_native),
-                errno=ER_INVALID_OCSP_RESPONSE)
+                errno=ER_INVALID_OCSP_RESPONSE_VALIDITY_INFO_MISSING)
 
         this_update = (this_update_native.replace(
             tzinfo=None) - SnowflakeOCSP.ZERO_EPOCH).total_seconds()
@@ -1397,7 +1491,7 @@ class SnowflakeOCSP(object):
             raise RevocationCheckError(
                 msg=SnowflakeOCSP._validity_error_message(
                     current_time, this_update, next_update),
-                errno=ER_INVALID_OCSP_RESPONSE)
+                errno=ER_OCSP_RESPONSE_EXPIRED)
 
     def _process_revoked_status(self, single_response, cert_id):
         """Processes REVOKED status."""
@@ -1415,7 +1509,7 @@ class SnowflakeOCSP(object):
                             SnowflakeOCSP.OUTPUT_TIMESTAMP_FORMAT,
                             gmtime(current_time)),
                         "Force Revoke"),
-                    errno=ER_SERVER_CERTIFICATE_REVOKED
+                    errno=ER_OCSP_RESPONSE_CERT_STATUS_REVOKED
                 )
 
         SnowflakeOCSP.OCSP_CACHE.delete_cache(self, cert_id)
@@ -1430,7 +1524,7 @@ class SnowflakeOCSP(object):
                 revocation_time.strftime(
                     SnowflakeOCSP.OUTPUT_TIMESTAMP_FORMAT),
                 revocation_reason),
-            errno=ER_SERVER_CERTIFICATE_REVOKED
+            errno=ER_OCSP_RESPONSE_CERT_STATUS_REVOKED
         )
 
     def _process_unknown_status(self, cert_id):
@@ -1438,7 +1532,7 @@ class SnowflakeOCSP(object):
         SnowflakeOCSP.OCSP_CACHE.delete_cache(self, cert_id)
         raise RevocationCheckError(
             msg="The certificate is in UNKNOWN revocation status.",
-            errno=ER_SERVER_CERTIFICATE_UNKNOWN,
+            errno=ER_OCSP_RESPONSE_CERT_STATUS_UNKNOWN,
         )
 
     def decode_ocsp_response_cache(self, ocsp_response_cache_json):
@@ -1453,7 +1547,9 @@ class SnowflakeOCSP(object):
                     self, cert_id, b64decode(ocsp_response), ts)
         except Exception as ex:
             logger.debug("Caught here - %s", ex)
-            raise ex
+            ermsg = "Exception raised while decoding OCSP Response Cache {}".format(str(ex))
+            raise RevocationCheckError(msg=ermsg,
+                                       errno=ER_OCSP_RESPONSE_CACHE_DECODE_FAILED)
 
     def encode_ocsp_response_cache(self, ocsp_response_cache_json):
         """Encodes OCSP response cache to JSON."""
